/* $Id: sflow.c,v 1.7 2002/06/14 13:54:04 aa5779 Exp $ */
/*
	Copyright (C) 2001 Artem V. Andreev (artem@AA5779.spb.edu)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include "interp.h"
#include "shared.h"
#include "flow.h"
#define SFLOW_IMP
#include "sflow.h"
#include "util.h"

flow_class script_flow_class = {"script", sf_open, sf_close, sf_getc, sf_ungetc, 
							   sf_putc, sf_puts,
							   sf_getpos, sf_setpos, sf_rewind,
							   sf_flowctl
							 };

typedef struct script_flow_data
{
	unsigned char writeable : 1;
	interpreter *filter;
	flow_class *nested_cls;
	flow_handle_t nested_data;
	int curch;
} script_flow_data;

static int out_int_data(script_flow_data *fd, dstring *result)
{
	char *iter;
	int bufch = 0;
	int flag = 0;
	int res;
	for(iter = DS_BODY(result); *iter; iter++)
	{
		if(isspace(*iter))
		{
			if(flag)
			{
				flag = 0;
				if((res = (fd->nested_cls->putc)(fd->nested_data,  bufch)))
					return res;
				bufch = 0;		
			}
		}
		else if(isdigit(*iter))
		{
			flag = 1;
			bufch = bufch * 10 + (*iter - '0');
		}
		else
		{
			if(flag)
			{
				flag = 0;
				if((res = (fd->nested_cls->putc)(fd->nested_data, bufch)))
					return res;
				bufch = 0;
			}
			if((res = (fd->nested_cls->putc)(fd->nested_data, *iter)))
				return res;
		}
	}
	return flag ? (fd->nested_cls->putc)(fd->nested_data, bufch) : 0;
}

flow_handle_t sf_open(char *name, char *options, int writeable)
{
	char *script_param = strchr(options, '?');
	script_flow_data *sfd = ALLOC(script_flow_data);
	dstring *retval;
	if(script_param)
		*script_param = '\0';
	sfd->writeable = writeable;
	sfd->filter = find_exported("subint", options);
	if(!sfd->filter)
	{
		FREE(sfd);
		recover_error("can't find exported interpreter '%s'", options);
	}
	retval = eval_procedurev(sfd->filter, "openflow", ds_fromint((int)sfd, 16, 0), 
								  ds_create(script_param), ds_create(writeable ? "true" : ""), 
								  ds_create(name),
								  NULL);
	if(retval)
	{
		ds_unfix(retval);
		if(*DS_BODY(retval))
			name = DS_BODY(retval);
	}
	sfd->nested_cls = detect_flow_class(name, &name, &options);
	sfd->nested_data = sfd->nested_cls->open(name, options, writeable);
	sfd->curch = -1;
	return sfd;
}

void sf_close(flow_handle_t fh)
{
	script_flow_data *sfd = fh;
	dstring *tmp;
	tmp = eval_procedurev(sfd->filter, "closeflow", ds_fromint((int)fh, 16, 0), NULL);
	if(tmp)
		ds_unfix(tmp);
	if(sfd->writeable && tmp && !ds_isempty(tmp))
		out_int_data(sfd, tmp);
	sfd->nested_cls->close(sfd->nested_data);
	FREE(sfd);
}

int sf_getc(flow_handle_t fh)
{
	script_flow_data *sfd = fh;
	dstring *tmp;
	variable *status_var = safe_get_var(sfd->filter, "keepit", 0);
	for(;;)
	{
		if(sfd->curch < 0)
		{
			sfd->curch = (sfd->nested_cls->getc)(sfd->nested_data);
			if(sfd->curch == FLOW_ERROR)
				return FLOW_ERROR;
		}
		tmp = eval_procedurev(sfd->filter, "getc", ds_fromint((int)fh, 16, 0), 
							 sfd->curch == FLOW_EOF ? ds_create("eof") : ds_createch(sfd->curch),
							 NULL);
		if(!status_var || ds_isempty(status_var->vals->val))
			sfd->curch = -1;
		if(tmp)
		{
			ds_unfix(tmp);
			if(!ds_isempty(tmp))
				break;
		}
	}
	if(!ds_comparestr(tmp, "eof", NULL, NULL))
		return FLOW_EOF;
	if(!ds_comparestr(tmp, "null", ds_p_casefold, NULL))
		return 0;
	return *DS_BODY(tmp);
}

int sf_ungetc(flow_handle_t fh, int ch)
{
	script_flow_data *sfd = fh;
	return (sfd->nested_cls->ungetc)(sfd->nested_data, ch);
}


int sf_putc(flow_handle_t fh, int ch)
{
	script_flow_data *sfd = fh;
	dstring *tmp;
	
	if(!sfd->writeable)
	{
		errno = EACCES;
		return FLOW_ERROR;
	}
	tmp = eval_procedurev(sfd->filter, "putc", ds_fromint((int)fh, 16, 0), 
						  ds_createch(ch), 
						  NULL);
	if(tmp)
	{
		ds_unfix(tmp);
		return out_int_data(sfd, tmp);
	}
	return 0;
}

int sf_puts(flow_handle_t fh, char *str)
{
	script_flow_data *sfd = fh;
	dstring *tmp;
	if(!str)
	{
		errno = EINVAL;
		return FLOW_ERROR;
	}
	if(!sfd->writeable)
	{
		errno = EACCES;
		return FLOW_ERROR;
	}
	tmp = eval_procedurev(sfd->filter, "puts", ds_fromint((int)fh, 16, 0), ds_create(str), NULL);
	if(tmp)
	{
		ds_unfix(tmp);
		return out_int_data(sfd, tmp);
	}
	return 0;
}
 
flow_pos_t sf_getpos(flow_handle_t fh)
{
	script_flow_data *sfd = fh;
	return sfd->nested_cls->getpos(sfd->nested_data);
}

int sf_setpos(flow_handle_t fh, flow_pos_t fp)
{
	script_flow_data *sfd = fh;
	return sfd->nested_cls->setpos(sfd->nested_data, fp);
}

int sf_rewind(flow_handle_t fh)
{
	script_flow_data *sfd = fh;
	return sfd->nested_cls->rewind(sfd->nested_data);
}

char *sf_flowctl(flow_handle_t fh, char *ctl)
{
	script_flow_data *sfd = fh;
	return sfd->nested_cls->flowctl(sfd->nested_data, ctl);
}
